local E, L = unpack(select(2, ...)) -- Engine, Locale
local CO, B = E:LoadModules("Config", "Blizzard")

--[[----------------------------------------------------

	This CUI Library provides a powerful toolset
	that is designed to create frame movers
	on the fly.
	
	You may use this Code in your own projects,
	as long as there is at least any credit given.
	
	Author: Ferodra / Arenima
	
----------------------------------------------------]]--

---------------------------------------------------
local _
local _G 						= _G
local pairs 					= pairs
local type 						= type
local HasExtraActionBar 		= HasExtraActionBar

local LibSticky					= LibStub("LibSimpleSticky-1.0")
---------------------------------------------------

E.Movers = {}
E.Stickys = {}

local MoverPanel_DescBase = '%s\n\n' .. L['MoverPanel_Description']

local function HideMoverPanel()
	E:UIFrameFadeOut(E.MoverPanel, 0.2, E.MoverPanel:GetAlpha(), 0)
	E.MoverPanel:EnableMouse(false)
end

local function MoverPanel_OnEnter(self)
	if E.HideTimer then
		E.HideTimer:Cancel()
	end
end

local function MoverPanel_OnLeave(self)
	if E.HideTimer then
		E.HideTimer:Cancel()
	end
	E.HideTimer = C_Timer.NewTimer(1, HideMoverPanel)
end

local function Panel_OnKeyDown(self, key)
	
	-- If a target key was pressed, continue execution
	-- If not, propagate input and continue normal execution for this key
	if key == 'UP' or key == 'DOWN' or key == 'LEFT' or key == 'RIGHT' then
		self:SetPropagateKeyboardInput(false)
		E:MoverPanel_Move(key)
	else
		self:SetPropagateKeyboardInput(true)
	end
end

local function Panel_OnEnter(self)
	self:EnableKeyboard(true)
	self:SetScript('OnKeyDown', Panel_OnKeyDown)
	
	MoverPanel_OnEnter()
end

local function Panel_OnLeave(self)
	self:EnableKeyboard(false)
	self:SetScript('OnKeyDown', nil)
	
	MoverPanel_OnLeave()
end

local SmartPositions = {
	["TOP"] = "BOTTOM",
	["BOTTOM"] = "TOP",
	["LEFT"] = "RIGHT",	
	["RIGHT"] = "LEFT",
	["TOPLEFT"] = "BOTTOMLEFT",
	["TOPRIGHT"] = "BOTTOMRIGHT",
	["BOTTOMLEFT"] = "TOPLEFT",
	["BOTTOMRIGHT"] = "TOPRIGHT",
	["CENTER"] = "CENTER",
}
function E:InversePosition(point)
	return SmartPositions[point]
end

function E:GetMoverPoints(Mover)
	local screenWidth, screenHeight, screenCenter = E.Parent:GetRight(), E.Parent:GetTop(), E.Parent:GetCenter()
	local x, y = Mover:GetCenter()
	
	local LEFT = screenWidth / 3
	local RIGHT = screenWidth * 2 / 3
	local TOP = screenHeight / 2
	local point, nudgePoint, nudgeInversePoint

	if y >= TOP then
		point = "TOP"
		nudgePoint = "TOP"
		y = -(screenHeight - Mover:GetTop())
	else
		point = "BOTTOM"
		nudgePoint = "BOTTOM"
		y = Mover:GetBottom()
	end

	if x >= RIGHT then
		point = point..'RIGHT'
		nudgePoint = "RIGHT"
		x = Mover:GetRight() - screenWidth
	elseif x <= LEFT then
		point = point..'LEFT'
		nudgePoint = "LEFT"
		x = Mover:GetLeft()
	else
		x = x - screenCenter
	end
	
	return x, y, point, nudgePoint
end

-- Toggle mover overlays and drag functionality
-- We can NOT simply show the whole thing, since we still have to hide them afterwards
-- this results in EVERY frame disappearing
function E:ToggleMover(state, noFade)
	
	-- Here we toggle special blizzard frames
	if state == true then
		-- Extra Button
		if not HasExtraActionBar() then
			ExtraActionBarFrame:Show()
			ExtraActionBarFrame:SetAlpha(1)
			ExtraActionButton1:Show()
		end
		
		VehicleSeatIndicator:Show()
	else
		-- Extra Button
		-- Prevent the button from being hidden when it actually is supposed to be active
		if not HasExtraActionBar() then
			ExtraActionBarFrame:Hide()
			ExtraActionBarFrame:SetAlpha(0)
			ExtraActionButton1:Hide()
		end
		
		VehicleSeatIndicator:Hide()
		
		if self.MoverPanel then
			self.MoverPanel:Hide()
		end
	end
	
	B:ToggleZoneAbility(state)
	
	-- We can assign "MoverChild.ForceMoverEnabled = true" to - force show it or false to force hide
	-- Assign nil to disable this functionality
	for k,v in pairs(self.Movers) do
		if state == true and (v.Frame.ForceMoverEnabled == true or (not v.Frame.ForceMoverEnabled and v.Frame.MoverEnabled)) then
			if v.Frame.ForceMoverEnabled == true or v.Frame.ForceMoverEnabled == false then
				if v.Frame.ForceMoverEnabled == true then
					v.Handle:EnableMouse(true)
					if not noFade then
						E:UIFrameFadeIn(v.Handle, 0.2, v.Handle:GetAlpha(), 1)
					else
						v.Handle:Show()
						v.Handle:SetAlpha(1)
					end
				else
					v.Handle:Hide()
					v.Handle:SetAlpha(0)
				end
			else
				if v.Frame.MoverEnabled then
					v.Handle:EnableMouse(true)
					if not noFade then
						E:UIFrameFadeIn(v.Handle, 0.2, v.Handle:GetAlpha(), 1)
					else
						v.Handle:Show()
						v.Handle:SetAlpha(1)
					end
				end
			end
		else
			v.Handle:EnableMouse(false)
			if not noFade then
				E:UIFrameFadeOut(v.Handle, 0.2, v.Handle:GetAlpha(), 0)
			else
				v.Handle:Hide()
				v.Handle:SetAlpha(0)
			end
		end
	end
end

function E:ApplyMoverConfig(Mover, Data)
	
	local Config = CO.db.profile.movers[Mover:GetName()]
	
	for k,v in pairs(Data) do
		if Config[k] and v then
			Config[k] = v
		end
	end
	
	self:LoadMoverPositions(Mover)
end

function E:GetMoverConfig(Mover)
	return CO.db.profile.movers[Mover:GetName()]
end

-- Returns a registered mover from child object or name-string
function E:GetMover(C)
	if type(C) == "string" then
		return self.Movers[C .. "Mover"]
	else
		return self.Movers[self:GetFullFrameName(C) .. "Mover"]
	end
end

function E:RegisterMover(M, MName)
	self.Movers[MName] = M
end

function E:LoadMoverPositions(limit)
	self = E
	
	local Conf = CO.db.profile.movers
	local ConfData = {}
	
	local SmartPosition
	
	if limit then
		local mover = self:GetMover(limit)
		
		if mover then
			ConfData = Conf[mover:GetName()]
			SmartPosition = ConfData["point"]
			
			if ConfData["enableAttach"] and ConfData["attachTo"] and ConfData["attachTo"][1] ~= "" then
				-- Attachment Target exists
				if _G[ConfData["attachTo"][1]] then
					mover:SetParent(_G[ConfData["attachTo"][1]])
					_G[ConfData["attachTo"][1]].Parent = mover
					mover.Frame.MoverEnabled = false
					
					SmartPosition = E:InversePosition(ConfData["point"])
				end
			else
				mover:SetParent(self.Parent)
				mover.Frame.MoverEnabled = true
				
				if ConfData["attachTo"] and ConfData["attachTo"][1] and _G[ConfData["attachTo"][1]] then
					_G[ConfData["attachTo"][1]].Parent = self.Parent					
				end
			end
			
			self:RepositionMover(mover, SmartPosition, ConfData["point"], ConfData["xOffset"] / mover:GetScale(), ConfData["yOffset"] / mover:GetScale())
			
			return
		end
	end
	
	-- k: Mover Name - v: Mover Object
	for k,v in pairs(self.Movers) do
		ConfData = Conf[k]
		
		if ConfData then
			if v then
				
				SmartPosition = ConfData["point"]
				
				if ConfData["enableAttach"] and ConfData["attachTo"] and ConfData["attachTo"][1] ~= "" then
					-- Attachment Target exists
					if _G[ConfData["attachTo"][1]] then
						v:SetParent(_G[ConfData["attachTo"][1]])
						_G[ConfData["attachTo"][1]].Parent = v
						v.Frame.MoverEnabled = false
						
						SmartPosition = E:InversePosition(ConfData["point"])
					end
				else
					v:SetParent(self.Parent)
					v.Frame.MoverEnabled = true
					
					if ConfData["attachTo"] and ConfData["attachTo"][1] and _G[ConfData["attachTo"][1]] then
						_G[ConfData["attachTo"][1]].Parent = self.Parent
					end
				end
				
				self:RepositionMover(v, SmartPosition, ConfData["point"], ConfData["xOffset"] / v:GetScale(), ConfData["yOffset"] / v:GetScale())
			else
				self:print("WARNING: Corrupt mover data found!")
			end
		end
	end
end

local function MoverHandle_OnDragStart(self)
	local parent = self:GetParent()
	
	parent:SetMovable(true)
	self:SetClampedToScreen(false)
	parent:SetClampedToScreen(false)
	
	if E.StickyMovers then
		LibSticky:StartMoving(parent, E.Stickys, E.StickyRange or 1, E.StickyRange or 1, E.StickyRange or 1, E.StickyRange or 1)
	else
		parent:StartMoving()
	end
end

local function MoverHandle_OnDragStop(self)
	local point, relativePoint, xOfs, yOfs
	local parent = self:GetParent()
	local title = self.Title
	
	parent:SetMovable(false)
	if E.StickyMovers then
		LibSticky:StopMoving(parent)
	else
		parent:StopMovingOrSizing()
	end
	
	-- Fix for movers without any default values
	if not CO.db.profile.movers[title] then
		CO.db.profile.movers[title] = {}
	end
	local conf = CO.db.profile.movers[title]
	
	
	if E.StickyMovers then
		xOfs, yOfs, point, relativePoint = E:GetMoverPoints(parent)
	else
		point, _, relativePoint, xOfs, yOfs = parent:GetPoint(parent:GetNumPoints())
	end
	
	conf["point"] 			= point
	conf["relativePoint"] 	= relativePoint
	conf["xOffset"] 		= xOfs
	conf["yOffset"] 		= yOfs
end

-- @PARAM1: Child target [Named frame(IMPORTANT), so we can store that config properly]
-- @PARAM2: Localized mover name for user display
-- @PARAM3: Alignment/Point
-- @PARAM4: Width
-- @PARAM5: Height
-- @PARAM6: Tooltip text for mover mode
function E:CreateMover(C, LT, A, X, Y, TT)
	local MNameRaw = self:GetFullFrameName(C)
	local MName = MNameRaw .. "Mover"
	
	if not A then A = "CENTER" end
	if not (X and Y) then X, Y = C:GetWidth(), C:GetHeight() if X <= 0 and Y <= 0 then X, Y = 50, 50 end end
	local M = self:NewFrame("Frame", MName, "LOW", X,Y, {"CENTER", self.Parent, "CENTER", 0, 0}, self.Parent)	
	
	-- Position update function
	-- This is probably the only existing way to deal with constantly moving blizzard frames and also work for everything else ofc
	local function Mover_SetPosition(_, _, parent)
		if parent ~= M then
			C:ClearAllPoints()
			C:SetParent(M)
			C:SetPoint(A, M, A, 0, 0)
		end
	end
	hooksecurefunc(C, "SetPoint", Mover_SetPosition)
	C:SetPoint("CENTER", self.Parent, "CENTER") -- Execute hook initially to force an update
	
	-- Create Mover handle to interact with
	M.Handle = self:CreateMoverHandle(C, LT, A, X, Y, TT)
	M.Handle:SetParent(M)
	M.Handle:SetFrameLevel(99)
	M.Handle:SetFrameStrata("HIGH")	
	
	-- Store reference to child frame
	C.MoverEnabled = true
	M.Frame = C
	
	-- E:GetMover(FrameObjectOrNameAsString).Handle:Show() -- We can use this to access the handle at any time!
	
	M.Handle.Title = MName
	M.Handle:SetScript("OnDragStart", MoverHandle_OnDragStart)
	M.Handle:SetScript("OnDragStop", MoverHandle_OnDragStop)
	
	-- Internal register
	-- MoverObject, MoverName
	self:RegisterMover(M, MName)
	
	return M
end

function E:RepositionMover(M, Point, RelativePoint, OffsetX, OffsetY)
	self:RepositionFrame(M, Point, RelativePoint, OffsetX, OffsetY)
end

-- Since we move frames via movers, which are basically frames, the frames we want to move with are parented to the mover
-- By later "copying" the translations made to the overlay to the base parent, we can move the whole thing
function E:CreateMoverHandle(C, LT, A, X, Y, TT)
	local MH = self:NewFrame("Frame", "Handle", "HIGH", X,Y, {A, C, A, 0, 0}, C)
	local RGB = self:GetUnitClassColor("player")
	MH:EnableMouse(true)
	MH:SetMovable(true)
	MH:RegisterForDrag("LeftButton")
	
	MH:SetBackdrop({
		bgFile 		= [[Interface\Buttons\WHITE8X8]],
		edgeFile 	= [[Interface\Buttons\WHITE8X8]],
		edgeSize 	= 1,
		tile 		= true, tileSize = 16
	})
	MH:SetBackdropColor(0.05, 0.05, 0.05, 0.6)
	MH:SetBackdropBorderColor(RGB[1], RGB[2], RGB[3], 0.8)
	
	-- Init font overlay
	MH.Name = MH:CreateFontString(nil, "ARTWORK")
	self:InitializeFontFrame(MH.Name, "ARTWORK", "FRIZQT__.TTF", 14, {1,1,1}, 1, {0,0}, "", 250, Y, MH, "CENTER", {1,1})
	MH.Name:SetFont(E.Media:Fetch("font", CO.db.profile.global.generalFont), 14)
	MH.Name:ClearAllPoints()
	MH.Name:SetPoint("CENTER", MH, 'CENTER', 0, 0)
	
	MH.Name:SetText(LT) -- Set provided name
	
	MH:SetAlpha(0) -- For toggle fading
	MH:Hide()
	
	-- Vars for tooltip functionality
	MH.LT = LT
	MH.TT = TT
	
	-- Add tooltip functionality to describe this mover in config mode
	MH:SetScript("OnEnter", self.Mover_OnEnter)
	MH:SetScript("OnLeave", self.Mover_OnLeave)
	
	table.insert(E.Stickys, MH)
	
	return MH
end

-- Update Mover dimensions based on child frame
function E:UpdateMoverDimensions(C)
	local M = self:GetMover(C)
	
	if not M then return false end
	
	M:SetSize(C:GetWidth(), C:GetHeight())
	M.Handle:SetSize(C:GetWidth(), C:GetHeight())
end

function E:Mover_OnEnter()
	
	Panel_OnEnter(self)
	MoverPanel_OnEnter()
	E:UpdateMoverPanel(self:GetParent())
	
	-- Set Tooltip
	E:Mover_GameTooltipOnEnter(self)
end

function E:Mover_OnLeave()
	
	Panel_OnEnter(self)
	MoverPanel_OnLeave()
	
	-- Set Normal State
	E:Mover_GameTooltipOnLeave(self)
end

-- TOOLTIP METHODS
	function E:Mover_GameTooltipOnEnter(self)
		-- If user actually wants tooltips
		if E.ShowMoverTooltips then
			GameTooltip:SetOwner(self, "ANCHOR_CURSOR")
			GameTooltip:AddLine(self.LT)
			if self.TT then
				GameTooltip:AddLine(self.TT)
			end
			GameTooltip:Show()
			
			-- Absolutely make sure it's on the very top!
			GameTooltip:SetFrameLevel(99999)
		end
		
		-- Highlight border to indicate the hovered frame
		self:SetBackdropBorderColor(0.7, 0.7, 0.7, 1)
	end

	function E:Mover_GameTooltipOnLeave(self)
		GameTooltip:Hide()
		local RGB = E:GetUnitClassColor("player")
		self:SetBackdropBorderColor(RGB[1], RGB[2], RGB[3], 0.8)
	end
-- TOOLTIP METHODS END

function E:ResetMoverPositions()
	CO.db.profile.movers = self:TableDeepCopy(self.MoverDefaults)
	
	self:LoadMoverPositions()
end

local function MoverPanel_Button_MoveUp()
	E:MoverPanel_Move('UP')
end

local function MoverPanel_Button_MoveDown()
	E:MoverPanel_Move('DOWN')
end

local function MoverPanel_Button_MoveLeft()
	E:MoverPanel_Move('LEFT')
end

local function MoverPanel_Button_MoveRight()
	E:MoverPanel_Move('RIGHT')
end

function E:MoverPanel_Move(direction)
	if not self.MoverPanel.Mover then return end
	
	local Panel = self.MoverPanel
	local Mover = Panel.Mover
	local Config = self:GetMoverConfig(Mover)
	local AxisX, AxisY
	
	if direction == 'UP' then
		AxisY = 1
	elseif direction == 'DOWN' then
		AxisY = -1
	elseif direction == 'LEFT' then
		AxisX = -1
	elseif direction == 'RIGHT' then
		AxisX = 1
	end
	
	if not AxisX and not AxisY then return end
	
	if AxisX then
		Config.xOffset = Config.xOffset + AxisX
	elseif AxisY then
		Config.yOffset = Config.yOffset + AxisY
	end
	
	E:ApplyMoverConfig(Mover, Config)
end

-- Repositions the Mover panel to an handle
function E:UpdateMoverPanel(Mover)
	
	-- Load mover panel only if we actually need it
	self:SetupMoverPanel()
	
	self.MoverPanel:ClearAllPoints()
	
	local PosX, PosY
	
	if Mover:GetLeft() > (GetScreenWidth() / 2) then
		PosX = 'LEFT'
	else
		PosX = 'RIGHT'
	end
	if Mover:GetTop() > (GetScreenHeight() / 2) then
		PosY = 'TOP'
	else
		PosY = 'BOTTOM'
	end
	
	local Pos = PosY .. PosX
	
	self.MoverPanel:SetPoint(Pos, Mover, self:InversePosition(Pos))
	
	E:UIFrameFadeIn(self.MoverPanel, 0.2, self.MoverPanel:GetAlpha(), 1)
	E.MoverPanel:EnableMouse(true)
	
	self.MoverPanel.Font.Text:SetText((MoverPanel_DescBase):format(Mover.Handle.LT or ''))
	-- Text
	
	-- Set Reference
	self.MoverPanel.Mover = Mover
end

function E:SetupMoverPanel()
	if not self.MoverPanel then
		self.MoverPanel = CreateFrame("Frame", "CUI_MoverPanel", self.Parent, 'InsetFrameTemplate')
		local Panel = self.MoverPanel
		
		Panel:SetSize(150, 200)
		Panel:SetFrameStrata("TOOLTIP")
		Panel:EnableMouse(true)
		Panel:SetClampedToScreen(true)
		--Panel.Background = self:CreateBackground(Panel)
		--Panel.Border = self:CreateBorder(Panel)
		Panel:Hide()
		
		Panel:SetScript('OnEnter', Panel_OnEnter)
		Panel:SetScript('OnLeave', Panel_OnLeave)
		
		--------------------------------
		
		Panel.Font = CreateFrame('Frame', nil, Panel)
		Panel.Font:SetPoint('TOPLEFT', Panel, 'TOPLEFT', 5, 0)
		Panel.Font:SetPoint('TOPRIGHT', Panel, 'TOPRIGHT', -5, 0)
		Panel.Font:SetHeight(125)
		
		Panel.Font.Text = E:CreateFont(Panel.Font)
		Panel.Font.Text:SetAllPoints(Panel.Font)
		
		--------------------------------
		
		local Buttons = {
			['Up'] = MoverPanel_Button_MoveUp,
			['Down'] = MoverPanel_Button_MoveDown,
			['Left'] = MoverPanel_Button_MoveLeft,
			['Right'] = MoverPanel_Button_MoveRight,
		}
		local NameBase, ChildBase = 'CUI_MoverPanel_Move%s', 'Button_%s'
		
		for Name, Func in pairs(Buttons) do
			local Button = CreateFrame('Button', (NameBase):format(Name), Panel, 'UIPanelSquareButton')
			Panel[(ChildBase):format(Name)] = Button
			
			Button:RegisterForClicks("AnyUp")
			Button:SetSize(25, 25)
			SquareButton_SetIcon(Button, string.upper(Name))
			Button:SetScript('OnClick', Func)
			Button:SetScript('OnEnter', Panel_OnEnter)
			Button:SetScript('OnLeave', Panel_OnLeave)
		end
		
		local Up, Down, Left, Right = Panel.Button_Up, Panel.Button_Down, Panel.Button_Left, Panel.Button_Right
		local ButtonOffsetY = -60
		
		Up:SetPoint('CENTER', Panel, 'CENTER', 0, ButtonOffsetY + (25))
		Down:SetPoint('CENTER', Panel, 'CENTER', 0, ButtonOffsetY + (-25))
		Left:SetPoint('CENTER', Panel, 'CENTER', -25, ButtonOffsetY + (0))
		Right:SetPoint('CENTER', Panel, 'CENTER', 25, ButtonOffsetY + (0))
	end
end